@@ -69,10 +69,6 @@ class Fitter(object):
     _REPLACE_AGE = 20
     _DEFAULT_SOURCE = {SOURCE.NAME: 'MOSFiT Paper'}
 
-    def __init__(self):
-        """Initialize `Fitter`."""
-        pass
-
     def fit_events(self,
                    events=[],
                    models=[],
@@ -82,10 +78,10 @@ class Fitter(object):
                    band_systems=[],
                    band_instruments=[],
                    band_bandsets=[],
-                   iterations=1000,
+                   iterations=5000,
                    num_walkers=None,
                    num_temps=1,
-                   parameter_paths=[''],
+                   parameter_paths=['parameters.json'],
                    fracking=True,
                    frack_step=50,
                    wrap_length=100,
@@ -103,11 +99,13 @@ class Fitter(object):
                    upload=False,
                    write=False,
                    quiet=False,
+                   cuda=False,
                    upload_token='',
                    check_upload_quality=False,
                    variance_for_each=[],
                    user_fixed_parameters=[],
-                   run_until_converged=False,
+                   convergence_type='psrf',
+                   convergence_criteria=None,
                    save_full_chain=False,
                    draw_above_likelihood=False,
                    maximum_walltime=False,
@@ -134,6 +132,15 @@ class Fitter(object):
         self._speak = speak
         self._limiting_magnitude = limiting_magnitude
 
+        self._cuda = cuda
+        if cuda:
+            try:
+                import pycuda.autoinit  # noqa: F401
+                import skcuda.linalg as linalg
+                linalg.init()
+            except ImportError:
+                pass
+
         dir_path = os.path.dirname(os.path.realpath(__file__))
         self._test = test
         self._wrap_length = wrap_length
@@ -188,8 +195,68 @@ class Fitter(object):
         ps = [[] for x in range(len(event_list))]
         lnprobs = [[] for x in range(len(event_list))]
 
+        # Load walker data if provided a list of walker paths.
         walker_data = []
 
+        if len(walker_paths):
+            try:
+                pool = MPIPool()
+            except ValueError:
+                pool = SerialPool()
+            if pool.is_master():
+                for walker_path in walker_paths:
+                    if os.path.exists(walker_path):
+                        with open(walker_path, 'r') as f:
+                            all_walker_data = json.load(
+                                f, object_pairs_hook=OrderedDict)
+                        prt.message('walker_file', [
+                                    walker_path], wrapped=True)
+
+                        # Support both the format where all data stored in a
+                        # single-item dictionary (the OAC format) and the older
+                        # MOSFiT format where the data was stored in the
+                        # top-level dictionary.
+                        if ENTRY.NAME not in all_walker_data:
+                            all_walker_data = all_walker_data[
+                                list(all_walker_data.keys())[0]]
+
+                        models = all_walker_data.get(ENTRY.MODELS, [])
+                        choice = None
+                        if len(models) > 1:
+                            model_opts = [
+                                '{}-{}-{}'.format(
+                                    x['code'], x['name'], x['date'])
+                                for x in models]
+                            choice = prt.prompt(
+                                'select_model_walkers', kind='select',
+                                message=True, options=model_opts)
+                            choice = model_opts.index(choice)
+                        elif len(models) == 1:
+                            choice = 0
+
+                        if choice is not None:
+                            walker_data.extend([
+                                x[REALIZATION.PARAMETERS]
+                                for x in models[choice][
+                                    MODEL.REALIZATIONS]])
+
+                        if not len(walker_data):
+                            prt.message('no_walker_data')
+                    else:
+                        prt.message('no_walker_data')
+                        if offline:
+                            prt.message('omit_offline')
+                        raise RuntimeError
+
+                for rank in range(1, pool.size + 1):
+                    pool.comm.send(walker_data, dest=rank, tag=3)
+            else:
+                walker_data = pool.comm.recv(source=0, tag=3)
+                pool.wait()
+
+            if pool.is_master():
+                pool.close()
+
         self._event_name = 'Batch'
         self._event_catalog = ''
         for ei, event in enumerate(event_list):
@@ -354,50 +421,10 @@ class Fitter(object):
                         pool.comm.send(self._event_name, dest=rank, tag=0)
                         pool.comm.send(path, dest=rank, tag=1)
                         pool.comm.send(data, dest=rank, tag=2)
-
-                    if len(walker_paths):
-                        walker_path = walker_paths[ei]
-                        if os.path.exists(walker_path):
-                            with open(walker_path, 'r') as f:
-                                all_walker_data = json.load(
-                                    f, object_pairs_hook=OrderedDict)
-                            prt.message('walker_file', [
-                                        walker_path], wrapped=True)
-                            models = all_walker_data.get(ENTRY.MODELS, [])
-                            choice = None
-                            if len(models) > 1:
-                                model_opts = [
-                                    '{}-{}-{}'.format(
-                                        x['code'], x['name'], x['date'])
-                                    for x in models]
-                                choice = prt.prompt(
-                                    'select_model_walkers', kind='select',
-                                    message=True, options=model_opts)
-                                choice = model_opts.index(choice)
-                            elif len(models) == 1:
-                                choice = 0
-
-                            if choice is not None:
-                                walker_data = [
-                                    x[REALIZATION.PARAMETERS]
-                                    for x in models[choice][
-                                        MODEL.REALIZATIONS]]
-
-                            if not len(walker_data):
-                                prt.message('no_walker_data')
-                        else:
-                            prt.message('no_walker_data')
-                            if offline:
-                                prt.message('omit_offline')
-                            raise RuntimeError
-
-                    for rank in range(1, pool.size + 1):
-                        pool.comm.send(walker_data, dest=rank, tag=3)
                 else:
                     self._event_name = pool.comm.recv(source=0, tag=0)
                     path = pool.comm.recv(source=0, tag=1)
                     data = pool.comm.recv(source=0, tag=2)
-                    walker_data = pool.comm.recv(source=0, tag=3)
                     pool.wait()
 
                 self._event_path = path
@@ -484,7 +511,8 @@ class Fitter(object):
                             upload=upload,
                             upload_token=upload_token,
                             check_upload_quality=check_upload_quality,
-                            run_until_converged=run_until_converged,
+                            convergence_type=convergence_type,
+                            convergence_criteria=convergence_criteria,
                             save_full_chain=save_full_chain,
                             extra_outputs=extra_outputs)
                         if return_fits:
@@ -690,7 +718,8 @@ class Fitter(object):
                  upload=False,
                  upload_token='',
                  check_upload_quality=True,
-                 run_until_converged=False,
+                 convergence_type='psrf',
+                 convergence_criteria=None,
                  save_full_chain=False,
                  extra_outputs=[]):
         """Fit the data for a given event.
@@ -826,12 +855,14 @@ class Fitter(object):
                     ntemps, nwalkers, ndim, likelihood, prior, pool=pool)
                 st = time.time()
             while (iterations > 0 and (
-                    run_until_converged or ici < len(iter_arr))):
+                    convergence_criteria is not None or ici < len(iter_arr))):
                 slr = int(np.round(sli))
-                ic = max_chunk if run_until_converged else iter_arr[ici]
+                ic = (max_chunk if convergence_criteria is not None else
+                      iter_arr[ici])
                 if exceeded_walltime:
                     break
-                if run_until_converged and converged and emi > iterations:
+                if (convergence_criteria is not None and converged and
+                        emi > iterations):
                     break
                 for li, (
                         p, lnprob, lnlike) in enumerate(
@@ -890,7 +921,12 @@ class Fitter(object):
                                     tar_x = np.array(
                                         p[np.random.randint(ntemps)][
                                             np.random.randint(nwalkers)])
-                                    new_x = np.clip(tar_x + dxx, 0.0, 1.0)
+                                    # Reflect if out of bounds.
+                                    new_x = np.clip(np.where(
+                                        np.where(tar_x + dxx < 1.0,
+                                                 tar_x + dxx,
+                                                 tar_x - dxx) > 0.0,
+                                        tar_x + dxx, tar_x - dxx), 0.0, 1.0)
                                     new_like = likelihood(new_x)
                                     new_prob = new_like + prior(new_x)
                                     if new_prob > wprob or np.isnan(wprob):
@@ -943,6 +979,17 @@ class Fitter(object):
                                 break
                         acor = [aacort, aa, ams]
 
+                        actc = int(np.ceil(aacort / sli))
+                        actn = np.int(float(emi - ams) / actc)
+
+                        if (convergence_type == 'acor' and
+                            convergence_criteria is not None and
+                            actn >= convergence_criteria and
+                                emi > iterations):
+                            prt.message('converged')
+                            converged = True
+                            break
+
                     # Calculate the PSRF (Gelman-Rubin statistic).
                     if li > 1 and emi > self._burn_in + 2:
                         cur_chain = (np.concatenate(
@@ -968,13 +1015,15 @@ class Fitter(object):
                         if np.isnan(psrf):
                             psrf = np.inf
 
-                        if (run_until_converged and psrf < 1.1 and
+                        if (convergence_type == 'psrf' and
+                            convergence_criteria is not None and
+                            psrf < convergence_criteria and
                                 emi > iterations):
                             prt.message('converged')
                             converged = True
                             break
 
-                    if run_until_converged:
+                    if convergence_criteria is not None:
                         self._emcee_est_t = -1.0
                     else:
                         self._emcee_est_t = float(
@@ -996,8 +1045,11 @@ class Fitter(object):
                             root='objective')
                         kmat = sout.get('kmat', None)
                         kdiag = sout.get('kdiagonal', None)
+                        variance = sout.get('variance')
                         if kdiag is not None and kmat is not None:
                             kmat[np.diag_indices_from(kmat)] += kdiag
+                        elif kdiag is not None and kmat is None:
+                            kmat = np.diag(kdiag + variance)
                     prt.status(
                         desc='fracking' if frack_now else
                         ('burning' if emi < self._burn_in else 'walking'),
@@ -1005,11 +1057,14 @@ class Fitter(object):
                         kmat=kmat,
                         accepts=accepts,
                         progress=[emi, None if
-                                  run_until_converged else iterations],
+                                  convergence_criteria is not None else
+                                  iterations],
                         acor=acor,
                         psrf=[psrf, self._burn_in],
                         messages=messages,
-                        make_space=emim1 == 0)
+                        make_space=emim1 == 0,
+                        convergence_type=convergence_type,
+                        convergence_criteria=convergence_criteria)
 
                     if s_exception:
                         break
@@ -1061,7 +1116,10 @@ class Fitter(object):
                         kmat=kmat,
                         fracking=True,
                         progress=[emi, None if
-                                  run_until_converged else iterations])
+                                  convergence_criteria is not None else
+                                  iterations],
+                        convergence_type=convergence_type,
+                        convergence_criteria=convergence_criteria)
                     tft = tft + time.time() - sft
                     if s_exception:
                         break
@@ -1138,14 +1196,13 @@ class Fitter(object):
         else:
             entry = Entry(name=self._event_name)
 
-        if upload:
-            uentry = Entry(name=self._event_name)
-            data_keys = set()
-            for task in model._call_stack:
-                if model._call_stack[task]['kind'] == 'data':
-                    data_keys.update(
-                        list(model._call_stack[task].get('keys', {}).keys()))
-            entryhash = entry.get_hash(keys=list(sorted(list(data_keys))))
+        uentry = Entry(name=self._event_name)
+        data_keys = set()
+        for task in model._call_stack:
+            if model._call_stack[task]['kind'] == 'data':
+                data_keys.update(
+                    list(model._call_stack[task].get('keys', {}).keys()))
+        entryhash = entry.get_hash(keys=list(sorted(list(data_keys))))
 
         # Accumulate all the sources and add them to each entry.
         sources = []
@@ -1155,13 +1212,12 @@ class Fitter(object):
         sources.append(entry.add_source(**self._DEFAULT_SOURCE))
         source = ','.join(sources)
 
-        if upload:
-            usources = []
-            if len(self._model._references):
-                for ref in self._model._references:
-                    usources.append(uentry.add_source(**ref))
-            usources.append(uentry.add_source(**self._DEFAULT_SOURCE))
-            usource = ','.join(usources)
+        usources = []
+        if len(self._model._references):
+            for ref in self._model._references:
+                usources.append(uentry.add_source(**ref))
+        usources.append(uentry.add_source(**self._DEFAULT_SOURCE))
+        usource = ','.join(usources)
 
         model_setup = OrderedDict()
         for ti, task in enumerate(model._call_stack):
@@ -1191,7 +1247,6 @@ class Fitter(object):
                     }
                 )
             if acor and aacort > 0:
-                actc = int(np.ceil(aacort))
                 acortimes = '<' if aa < self._MAX_ACORC else ''
                 acortimes += str(np.int(float(emi - ams) / actc))
                 modeldict[MODEL.CONVERGENCE].append(
@@ -1202,19 +1257,19 @@ class Fitter(object):
                 )
             modeldict[MODEL.STEPS] = str(emi)
 
-        if upload:
-            umodeldict = deepcopy(modeldict)
-            umodeldict[MODEL.SOURCE] = usource
-            modelhash = get_model_hash(
-                umodeldict, ignore_keys=[MODEL.DATE, MODEL.SOURCE])
-            umodelnum = uentry.add_model(**umodeldict)
-            if check_upload_quality:
-                if WAIC is None:
-                    upload_model = False
-                elif WAIC is not None and WAIC < 0.0:
+        umodeldict = deepcopy(modeldict)
+        umodeldict[MODEL.SOURCE] = usource
+        modelhash = get_model_hash(
+            umodeldict, ignore_keys=[MODEL.DATE, MODEL.SOURCE])
+        umodelnum = uentry.add_model(**umodeldict)
+        if check_upload_quality:
+            if WAIC is None:
+                upload_model = False
+            elif WAIC is not None and WAIC < 0.0:
+                if upload:
                     prt.message('no_ul_waic', ['' if WAIC is None
                                                else pretty_num(WAIC)])
-                    upload_model = False
+                upload_model = False
 
         modelnum = entry.add_model(**modeldict)
 
@@ -1232,7 +1287,6 @@ class Fitter(object):
         # upon the value of acort (the autocorrelation timescale).
         if acor and aacort > 0 and aa == self._MAX_ACORC:
             actc0 = int(np.ceil(aacort))
-            actc = int(np.ceil(aacort / sli))
             for i in range(1, np.int(float(emi - ams) / actc0)):
                 pout = np.concatenate(
                     (all_chain[:, :, -i * actc, :], pout), axis=1)
@@ -1343,13 +1397,12 @@ class Fitter(object):
                             compare_to_existing=False, check_for_dupes=False,
                             **photodict)
 
-                        if upload_model:
-                            uphotodict = deepcopy(photodict)
-                            uphotodict[PHOTOMETRY.SOURCE] = umodelnum
-                            uentry.add_photometry(
-                                compare_to_existing=False,
-                                check_for_dupes=False,
-                                **uphotodict)
+                        uphotodict = deepcopy(photodict)
+                        uphotodict[PHOTOMETRY.SOURCE] = umodelnum
+                        uentry.add_photometry(
+                            compare_to_existing=False,
+                            check_for_dupes=False,
+                            **uphotodict)
                 else:
                     output = model.run_stack(y, root='objective')
 
@@ -1384,19 +1437,23 @@ class Fitter(object):
                 realdict[REALIZATION.ALIAS] = str(ri)
                 entry[ENTRY.MODELS][0].add_realization(**realdict)
                 urealdict = deepcopy(realdict)
-                if upload_model:
-                    uentry[ENTRY.MODELS][0].add_realization(**urealdict)
+                uentry[ENTRY.MODELS][0].add_realization(**urealdict)
                 ri = ri + 1
         prt.message('all_walkers_written', inline=True)
 
         entry.sanitize()
-        oentry = entry._ordered(entry)
+        oentry = {self._event_name: entry._ordered(entry)}
+        uentry.sanitize()
+        ouentry = {self._event_name: uentry._ordered(uentry)}
+
+        uname = '_'.join(
+            [self._event_name, entryhash, modelhash])
 
         if not os.path.exists(model.MODEL_OUTPUT_DIR):
             os.makedirs(model.MODEL_OUTPUT_DIR)
 
         if write:
-            prt.message('writing_model')
+            prt.message('writing_complete')
             with open_atomic(
                     os.path.join(model.MODEL_OUTPUT_DIR, 'walkers.json'),
                     'w') as flast, open_atomic(os.path.join(
@@ -1432,18 +1489,24 @@ class Fitter(object):
                     entabbed_json_dump(extras, flast, separators=(',', ':'))
                     entabbed_json_dump(extras, feven, separators=(',', ':'))
 
+            prt.message('writing_model')
+            with open_atomic(os.path.join(
+                model.MODEL_OUTPUT_DIR, 'upload.json'),
+                    'w') as flast, open_atomic(os.path.join(
+                        model.MODEL_OUTPUT_DIR,
+                        uname + (('_' + suffix) if suffix else '') +
+                        '.json'), 'w') as feven:
+                entabbed_json_dump(ouentry, flast, separators=(',', ':'))
+                entabbed_json_dump(ouentry, feven, separators=(',', ':'))
+
         if upload_model:
-            uentry.sanitize()
             prt.message('ul_fit', [entryhash, modelhash])
-            upath = '/' + '_'.join(
-                [self._event_name, entryhash, modelhash]) + '.json'
-            ouentry = {self._event_name: uentry._ordered(uentry)}
             upayload = entabbed_json_dumps(ouentry, separators=(',', ':'))
             try:
                 dbx = dropbox.Dropbox(upload_token)
                 dbx.files_upload(
                     upayload.encode(),
-                    upath,
+                    '/' + uname + '.json',
                     mode=dropbox.files.WriteMode.overwrite)
                 prt.message('ul_complete')
             except Exception:
