@@ -92,6 +92,7 @@ class Photometry(Module):
         self._band_areas = [[] for i in range(self._n_bands)]
         self._min_waves = np.full(self._n_bands, 0.0)
         self._max_waves = np.full(self._n_bands, 0.0)
+        self._imp_waves = [[0.0, 1.0] for i in range(self._n_bands)]
         self._filter_integrals = np.full(self._n_bands, 0.0)
         self._average_wavelengths = np.full(self._n_bands, 0.0)
         self._band_offsets = np.full(self._n_bands, 0.0)
@@ -100,6 +101,7 @@ class Photometry(Module):
         self._band_xu = np.full(self._n_bands, u.Angstrom.cgs.scale)
         self._band_yu = np.full(self._n_bands, 1.0)
         self._band_kinds = np.full(self._n_bands, 'magnitude', dtype=object)
+        self._band_index_cache = {}
 
         for i, band in enumerate(self._unique_bands):
             self._band_xunits[i] = band.get('xunit', 'Angstrom')
@@ -116,7 +118,6 @@ class Photometry(Module):
 
         if self._pool.is_master():
             vo_tabs = OrderedDict()
-            prt.prt('')
 
         per = 0.0
         bc = 0
@@ -234,7 +235,7 @@ class Photometry(Module):
                 xscale = (c.h * c.c /
                           u.Angstrom).cgs.value / self._band_xu[i]
                 self._band_energies[
-                    i], self._band_areas[i] = xvals, yvals
+                    i], self._band_areas[i] = xvals, yvals / xvals
                 self._band_wavelengths[i] = xscale / self._band_energies[i]
                 self._average_wavelengths[i] = np.trapz([
                     x * y
@@ -271,6 +272,15 @@ class Photometry(Module):
 
             self._min_waves[i] = min(self._band_wavelengths[i])
             self._max_waves[i] = max(self._band_wavelengths[i])
+            self._imp_waves[i] = [self._min_waves[i], self._max_waves[i]]
+            if len(self._transmissions[i]):
+                new_wave = self._band_wavelengths[i][
+                    np.argmax(self._transmissions[i])]
+                self._imp_waves[i].append(new_wave)
+            elif len(self._band_areas[i]):
+                new_wave = self._band_wavelengths[i][
+                    np.argmax(self._band_areas[i])]
+                self._imp_waves[i].append(new_wave)
             bc = bc + 1
 
         if self._pool.is_master():
@@ -282,6 +292,10 @@ class Photometry(Module):
         """Find the index corresponding to the provided band information."""
         bmatch = 0
         bbi = None
+        cache_key = ':'.join([
+            band, telescope, instrument, mode, bandset, system])
+        if cache_key in self._band_index_cache:
+            return self._band_index_cache[cache_key]
         for bi, bnd in enumerate(self._unique_bands):
             matches = [band == bnd['name'],
                        system == self._band_systs[bi],
@@ -303,6 +317,7 @@ class Photometry(Module):
             if lmatch == 6 and nbmatch > 0:
                 break
         if bbi is not None:
+            self._band_index_cache[cache_key] = bbi
             return bbi
         raise ValueError(
             'Cannot find band index for `{}` band of bandset `{}` '
@@ -336,18 +351,17 @@ class Photometry(Module):
                 if self._observation_types[li] == 'magnitude':
                     offsets[li] = self._band_offsets[bi]
                     wavs = kwargs['sample_wavelengths'][bi]
-                    dx = wavs[1] - wavs[0]
                     yvals = np.interp(
                         wavs, self._band_wavelengths[bi],
                         self._transmissions[bi]) * kwargs['seds'][li] / zp1
                     eff_fluxes[li] = np.trapz(
-                        yvals, dx=dx) / self._filter_integrals[bi]
+                        yvals, wavs) / self._filter_integrals[bi]
                 elif self._observation_types[li] == 'countrate':
                     wavs = np.array(kwargs['sample_wavelengths'][bi])
                     yvals = (np.interp(
                         wavs, self._band_wavelengths[bi], self._band_areas[
                             bi]) * kwargs['seds'][li] / zp1 / (
-                                self.H_C_ANG_CGS / wavs))
+                                self.H_C_ANG_CGS / wavs) / self.ANG_CGS)
                     eff_fluxes[li] = np.trapz(yvals, wavs)
                 else:
                     raise RuntimeError('Unknown observation kind.')
@@ -419,7 +433,7 @@ class Photometry(Module):
         if request == 'photometry':
             return self
         elif request == 'band_wave_ranges':
-            return list(map(list, zip(*[self._min_waves, self._max_waves])))
+            return self._imp_waves
         elif request == 'average_wavelengths':
             return self._average_wavelengths
         elif request == 'variance_bands':
